;; Problem 59
;; 19 December 2003
;;
;; Each character on a computer is assigned a unique code and the
;; preferred standard is ASCII (American Standard Code for Information
;; Interchange). For example, uppercase A = 65, asterisk (*) = 42, and
;; lowercase k = 107.
;;
;; A modern encryption method is to take a text file, convert the
;; bytes to ASCII, then XOR each byte with a given value, taken from a
;; secret key. The advantage with the XOR function is that using the
;; same encryption key on the cipher text, restores the plain text;
;; for example, 65 XOR 42 = 107, then 107 XOR 42 = 65.
;;
;; For unbreakable encryption, the key is the same length as the plain
;; text message, and the key is made up of random bytes. The user
;; would keep the encrypted message and the encryption key in
;; different locations, and without both "halves", it is impossible to
;; decrypt the message.
;;
;; Unfortunately, this method is impractical for most users, so the
;; modified method is to use a password as a key. If the password is
;; shorter than the message, which is likely, the key is repeated
;; cyclically throughout the message. The balance for this method is
;; using a sufficiently long password key for security, but short
;; enough to be memorable.
;;
;; Your task has been made easy, as the encryption key consists of
;; three lower case characters. Using cipher1.txt (right click and
;; 'Save Link/Target As...'), a file containing the encrypted ASCII
;; codes, and the knowledge that the plain text must contain common
;; English words, decrypt the message and find the sum of the ASCII
;; values in the original text.

(require 'asdf)
(require 'cl-ppcre)

(defvar *fil-cont*
  (with-open-file (inp "cipher1.txt" :direction :input)
    (read-line inp)))

(defvar *doc-characters*
  (mapcar #'read-from-string (cl-ppcre:split "," *fil-cont*)))

(defun histogram (lst)
  (let ((ht (make-hash-table :test #'equal)))
    (loop for ch in lst
	  do
	  ;(format t "ch:~a : ~a~&" ch (gethash ch ht))
	  (when (not (gethash ch ht))
	    (setf (gethash ch ht) 0))
	  (setf (gethash ch ht)
		(+ 1 (gethash ch ht))))
    (loop for ch being the hash-keys of ht
	  collect
	  (list ch (gethash ch ht)))))

; (histogram *doc-characters*)

;; (sort (histogram *doc-characters*)
;;       #'>
;;       :key #'second)

(defun group-by (lst num-items)
  (labels ((inner (acc items num-items)
	     (cond 
	       ((= 0 num-items)
		(reverse acc))
	       ((null items)
		(reverse acc))
	       (t
		(let ((end-idx (min (length items) num-items)))
		  (inner (cons (subseq items 0 end-idx) acc)
			 (subseq items end-idx)
			 end-idx))))))
    (inner (list) lst num-items)))

;(group-by '(1 2 3 4 5 6 7) 3)

(defvar *grouped-by-3*
  (group-by *doc-characters* 3))


(format t "most common 'a':~a~&"
	(first 
	 (sort (histogram (mapcar #'first *grouped-by-3*))
	       #'>
	       :key #'second)))

;(format t "~b~&" 71)
;(format t "~b~&" 32)
;1000111
; 100000
;#b0100111

(format t "most common 'b':~a~&"
	(first 
	 (sort (histogram (mapcar #'second *grouped-by-3*))
	       #'>
	       :key #'second)))

;(format t "~b~&~b~&" 79 32)
;  1001111
;   100000
;#b1101111


(format t "most common 'c':~a~&"
	(first 
	 (sort (histogram (mapcar #'third *grouped-by-3*))
	       #'>
	       :key #'second)))

;(format t "~b~&~b~&" 68 32)

;  1000100
;  0100000
;#b1100100


;; 'a' == 97
(defun decrypt (pass doc)
  (loop for echr in doc
	for pass-idx = 0 then (mod (+ pass-idx 1) (length pass))
	do
	(format t "~a" (code-char (logxor (nth pass-idx pass) echr)))))

(defun decrypt->string (pass doc)
  (with-output-to-string (out)
    (loop for echr in doc
	  for pass-idx = 0 then (mod (+ pass-idx 1) (length pass))
	  do
	  (format out "~a" (code-char (logxor (nth pass-idx pass) echr))))))

; (decrypt '(39 111 100) *doc-characters*)
(loop
 for chr1 from 1 to 127
 do
 (format t "chr1:~a~&" chr1)
 (decrypt (list chr1 111 100) *doc-characters*)
 (format t "~&"))

;; (decrypt->string '(103 111 100) *doc-characters*)

(apply #'+ (mapcar #'char-code
		   (mapcar #'(lambda (str) (char str 0))
			   (cl-ppcre:split "" (decrypt->string '(103 111 100) *doc-characters*)))))


;(format t "~b" (char-code #\%))
;(format t "~b" (char-code #\e))
;(format t "~b" 39)
;; 1100100
;; 0100101

;; 1000001



;(expt 127 3)